/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 2017                                                     **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#include <config.h>

#include "MutableGlobalDefs.h"

#include <gtest/gtest.h>

#include <pearl/CallingContext.h>
#include <pearl/Callsite.h>
#include <pearl/Cartesian.h>
#include <pearl/Communicator.h>
#include <pearl/Error.h>
#include <pearl/Location.h>
#include <pearl/LocationGroup.h>
#include <pearl/Metric.h>
#include <pearl/ProcessGroup.h>
#include <pearl/Region.h>
#include <pearl/RmaWindow.h>
#include <pearl/SourceLocation.h>
#include <pearl/String.h>
#include <pearl/SystemNode.h>

#include "Process.h"

using namespace testing;
using namespace pearl;
using namespace pearl::detail;


// --- Helper ---------------------------------------------------------------

namespace
{
// Dummy test fixture needed for MutableGlobalDefs typed tests
template< typename DefClassT >
class MutableGlobalDefsT
    : public Test
{
};


// Instantiate test cases for definition types supported by MutableGlobalDefs
typedef Types< CallingContext,
               Callsite,
               Cartesian,
               Communicator,
               Location,
               LocationGroup,
               Metric,
               ProcessGroup,
               Region,
               RmaWindow,
               SourceLocation,
               String,
               SystemNode > DefinitionTypes;
TYPED_TEST_CASE(MutableGlobalDefsT, DefinitionTypes);


// createItem function template and specializations for supported definitions
template< typename DefClassT >
DefClassT*
createItem(const typename DefClassT::IdType id)
{
    return 0;
}


template< >
CallingContext*
createItem(const CallingContext::IdType id)
{
    return new CallingContext(id,
                              Region::UNDEFINED,
                              SourceLocation::UNDEFINED,
                              0);
}


template< >
Callsite*
createItem(const Callsite::IdType id)
{
    return new Callsite(id, String::UNDEFINED, 0, Region::UNDEFINED);
}


template< >
Cartesian*
createItem(const Cartesian::IdType id)
{
    Cartesian::cart_dims   dims;
    Cartesian::cart_period period;

    return new Cartesian(id, dims, period);
}


template< >
Communicator*
createItem(const Communicator::IdType id)
{
    return new Communicator(id,
                            String::UNDEFINED,
                            ProcessGroup::UNDEFINED,
                            Paradigm::UNKNOWN,
                            0);
}


template< >
Location*
createItem(const Location::IdType id)
{
    Process process(0,
                    String::UNDEFINED,
                    const_cast< SystemNode* >(&SystemNode::UNDEFINED));

    return new Location(id,
                        String::UNDEFINED,
                        Location::TYPE_CPU_THREAD,
                        0,
                        &process);
}


template< >
LocationGroup*
createItem(const LocationGroup::IdType id)
{
    return new Process(id,
                       String::UNDEFINED,
                       const_cast< SystemNode* >(&SystemNode::UNDEFINED));
}


template< >
Metric*
createItem(const Metric::IdType id)
{
    return new Metric(id, "", "", "INTEGER", "COUNTER", "START");
}


template< >
ProcessGroup*
createItem(const ProcessGroup::IdType id)
{
    ProcessGroup::RankContainer ranks;

    return new ProcessGroup(id, String::UNDEFINED, ranks, true, true);
}


template< >
Region*
createItem(const Region::IdType id)
{
    return new Region(id,
                      String::UNDEFINED,
                      String::UNDEFINED,
                      String::UNDEFINED,
                      Region::ROLE_UNKNOWN,
                      Paradigm::UNKNOWN,
                      String::UNDEFINED,
                      0,
                      0);
}


template< >
RmaWindow*
createItem(const RmaWindow::IdType id)
{
    return new RmaWindow(id,
                         const_cast< Communicator* >(&Communicator::UNDEFINED));
}


template< >
SourceLocation*
createItem(const SourceLocation::IdType id)
{
    return new SourceLocation(id, String::UNDEFINED, 0);
}


template< >
String*
createItem(const String::IdType id)
{
    return new String(id, "");
}


template< >
SystemNode*
createItem(const SystemNode::IdType id)
{
    return new SystemNode(id, String::UNDEFINED, String::UNDEFINED, 0);
}


// addItem function template and specializations for supported definitions
template< typename DefClassT >
void
addItem(MutableGlobalDefs& defs,
        DefClassT*         item)
{
}


#define ADDITEM_FUNC(DefClassT)       \
    template< >                       \
    void                              \
    addItem(MutableGlobalDefs & defs, \
            DefClassT * item)         \
    {                                 \
        defs.add ## DefClassT(item);  \
    }

ADDITEM_FUNC(CallingContext)
ADDITEM_FUNC(Callsite)
ADDITEM_FUNC(Cartesian)
ADDITEM_FUNC(Communicator)
ADDITEM_FUNC(Location)
ADDITEM_FUNC(LocationGroup)
ADDITEM_FUNC(Metric)
ADDITEM_FUNC(ProcessGroup)
ADDITEM_FUNC(Region)
ADDITEM_FUNC(RmaWindow)
ADDITEM_FUNC(SourceLocation)
ADDITEM_FUNC(String)
ADDITEM_FUNC(SystemNode)

#undef ADDITEM_FUNC
}    // unnamed namespace


// --- MutableGlobalDefs tests ----------------------------------------------

TYPED_TEST(MutableGlobalDefsT, validate_defsWithUniqueIds_doesNotThrow)
{
    MutableGlobalDefs defs;

    addItem< TypeParam >(defs, createItem< TypeParam >(7));
    addItem< TypeParam >(defs, createItem< TypeParam >(2));
    addItem< TypeParam >(defs, createItem< TypeParam >(5));
    addItem< TypeParam >(defs, createItem< TypeParam >(3));

    EXPECT_NO_THROW(defs.validate());
}


TYPED_TEST(MutableGlobalDefsT, validate_defsWithDuplicateIds_throws)
{
    MutableGlobalDefs defs;

    addItem< TypeParam >(defs, createItem< TypeParam >(7));
    addItem< TypeParam >(defs, createItem< TypeParam >(2));
    addItem< TypeParam >(defs, createItem< TypeParam >(5));
    addItem< TypeParam >(defs, createItem< TypeParam >(2));

    EXPECT_THROW(defs.validate(), RuntimeError);
}
